#!/usr/bin/env lua

require 'Test.Assertion'

plan(70)

local mqtt = require 'mqtt311'

local sock = {
    close = true,
    read  = true,
    write = function (_, s) return { string.byte(s, 1, #s) } end,
}
local cli = mqtt({
    socket = sock,
    logger = {
        debug = function (_, ...) diag(...) end,
        error = function (_, ...) diag(...) end,
        fatal = function (_, ...) diag(...) end,
        info  = function (_, ...) diag(...) end,
        warn  = function (_, ...) diag(...) end,
    },
})
is_table( cli )

local res = cli:connect({ clean = true })
equals( res[1], 0x10 )
equals( res[10], 0x02, "clean session" )

res = cli:reconnect(sock)
equals( res[1], 0x10 )
equals( res[10], 0x02 )

cli.connect_sent = false
res = cli:connect({ keep_alive = 42, id = 'CLI', username = 'user', password = 'pAssWd', will = {'a/b', 'ended', qos = 2, retain = true } })
equals( res[1], 0x10 )
equals( res[10], 0xF4 )

res = cli:reconnect(sock)
equals( res[1], 0x10 )
equals( res[10], 0xF4 )

cli.connect_sent = false
res = cli:connect({ keep_alive = 42, id = 'CLI', username = 'user', password = 'pAssWd', will = {'a/b', 'ended' } })
equals( res[1], 0x10 )
equals( res[10], 0xC4 )

res = cli:reconnect(sock)
equals( res[1], 0x10 )
equals( res[10], 0xC4 )

cli.connect_sent = false
error_matches( function () cli:connect(42) end,
               "^[^:]+:%d+: bad argument #1 to 'connect' %(table expected, got number%)" )

error_matches( function () cli:connect({ id = 'foo', keep_alive = 3.14 }) end,
               "^[^:]+:%d+: invalid Keep Alive %(3.14%) in table to 'connect'" )

error_matches( function () cli:connect({ id = 'foo', keep_alive = 1000000 }) end,
               "^[^:]+:%d+: invalid Keep Alive %(1000000%) in table to 'connect'" )

error_matches( function () cli:connect({ id = 42 }) end,
               "^[^:]+:%d+: invalid Client Identifier %(42%) in table to 'connect'" )

error_matches( function () cli:connect({ id = 'foo', username = 42 }) end,
               "^[^:]+:%d+: invalid User Name %(42%) in table to 'connect'" )

error_matches( function () cli:connect({ id = 'foo', password = 42 }) end,
               "^[^:]+:%d+: invalid Password %(42%) in table to 'connect'" )

error_matches( function () cli:connect({ id = 'foo', password = 'pAssWd' }) end,
               "^[^:]+:%d+: Password without User Name in table to 'connect'" )

error_matches( function () cli:connect({ id = 'foo', will = 42 }) end,
               "^[^:]+:%d+: invalid Will %(42%) in table to 'connect'" )

error_matches( function () cli:connect({ id = 'foo', will = {'a\0b', 'ended' } }) end,
               "^[^:]+:%d+: invalid Will Topic %(a\0b%) in table to 'connect'" )

error_matches( function () cli:connect({ id = 'foo', will = {'a/b', 42 } }) end,
               "^[^:]+:%d+: invalid Will Message %(42%) in table to 'connect'" )

error_matches( function () cli:connect({ id = 'foo', will = {'a/b', 'ended', qos = 42 } }) end,
               "^[^:]+:%d+: invalid Will QoS %(42%) in table to 'connect'" )

cli.connect_sent = true
error_matches( function () cli:reconnect(42) end,
               "^[^:]+:%d+: bad argument #1 to 'reconnect' %(socket expected, got number%)" )

truthy(cli:reconnect(sock))

res = cli:publish('a/b', 'FooBar')
equals( res[1], 0x30 )

res = cli:publish('a/b', 'FooBar', { qos = 1 })
equals( res[1], 0x32 )

res = cli:publish('a/b', 'FooBar', { qos = 1, retain = true })
equals( res[1], 0x33 )

res = cli:publish('a/b', 'FooBar', { qos = 2 })
equals( res[1], 0x34 )

res = cli:publish('a/b', 'FooBar', { qos = 2, retain = true })
equals( res[1], 0x35 )

error_matches( function () cli:publish(42, 'FooBar') end,
               "^[^:]+:%d+: bad argument #1 to 'match' %(string expected, got number%)" )

error_matches( function () cli:publish('', 'FooBar') end,
               "^[^:]+:%d+: invalid Topic Name to 'publish'" )

error_matches( function () cli:publish('\xFF', 'FooBar') end,
               "^[^:]+:%d+: invalid Topic Name to 'publish'" )

error_matches( function () cli:publish('a\0b', 'FooBar') end,
               "^[^:]+:%d+: invalid Topic Name to 'publish'" )

error_matches( function () cli:publish('a/#', 'FooBar') end,
               "^[^:]+:%d+: invalid Topic Name to 'publish'" )

error_matches( function () cli:publish('+/b', 'FooBar') end,
               "^[^:]+:%d+: invalid Topic Name to 'publish'" )

error_matches( function () cli:publish('$a/b', 'FooBar') end,
               "^[^:]+:%d+: invalid Topic Name to 'publish'" )

error_matches( function () cli:publish('a/b', 42) end,
               "^[^:]+:%d+: bad argument #2 to 'publish' %(string expected, got number%)" )

error_matches( function () cli:publish('a/b', 'FooBar', 42 ) end,
               "^[^:]+:%d+: bad argument #3 to 'publish' %(table expected, got number%)" )

error_matches( function () cli:publish('a/b', 'FooBar', { qos = 42 }) end,
               "^[^:]+:%d+: invalid QoS %(42%) in table to 'publish'" )

res = cli:subscribe({ 'a/b', 0x1, 'c/d', 0x2 })
equals( res[1], 0x82 )

error_matches( function () cli:subscribe(42) end,
               "^[^:]+:%d+: bad argument #1 to 'subscribe' %(table expected, got number%)" )

error_matches( function () cli:subscribe({}) end,
               "^[^:]+:%d+: empty table for 'subscribe'" )

error_matches( function () cli:subscribe({ 'a/b', 0x1, '', 0x2 }) end,
               "^[^:]+:%d+: invalid Topic Filter %(%) at index 3 in table to 'subscribe'" )

error_matches( function () cli:subscribe({ 'a/b', 0x1, '\xFF', 0x2 }) end,
               "^[^:]+:%d+: invalid Topic Filter %(\xFF%) at index 3 in table to 'subscribe'" )

error_matches( function () cli:subscribe({ 'a/b', 0x1, 'c\0d', 0x2 }) end,
               "^[^:]+:%d+: invalid Topic Filter %(c\0d%) at index 3 in table to 'subscribe'" )

error_matches( function () cli:subscribe({ 'a/b', 0x1, 42, 0x2 }) end,
               "^[^:]+:%d+: invalid Topic Filter %(42%) at index 3 in table to 'subscribe'" )

error_matches( function () cli:subscribe({ 'a/b', 0x1, 'c/d', 42 }) end,
               "^[^:]+:%d+: invalid Requested QoS %(42%) at index 4 in table to 'subscribe'" )

res = cli:unsubscribe({ 'a/b', 'c/d' })
equals( res[1], 0xA2 )

error_matches( function () cli:unsubscribe(42) end,
               "^[^:]+:%d+: bad argument #1 to 'unsubscribe' %(table expected, got number%)" )

error_matches( function () cli:unsubscribe({}) end,
               "^[^:]+:%d+: empty table for 'unsubscribe'" )

error_matches( function () cli:unsubscribe({'a/b', 'c\0d'}) end,
               "^[^:]+:%d+: invalid Topic Filter %(c\0d%) at index 2 in table to 'unsubscribe'" )

error_matches( function () cli:unsubscribe({'a/b', 42}) end,
               "^[^:]+:%d+: invalid Topic Filter %(42%) at index 2 in table to 'unsubscribe'" )

res = cli:ping()
equals( res[1], 0xC0 )

res = cli:disconnect()
equals( res[1], 0xE0 )


cli = mqtt({
    socket = sock,
})

error_matches( function () cli:reconnect(sock) end,
               "^[^:]+:%d+: CONNECT not sent" )

error_matches( function () cli:publish('a/b', 'FooBar') end,
               "^[^:]+:%d+: CONNECT not sent" )

error_matches( function () cli:subscribe({ 'a/b', 0x1, 'c/d', 0x2 }) end,
               "^[^:]+:%d+: CONNECT not sent" )

error_matches( function () cli:unsubscribe({ 'a/b', 'c/d' }) end,
               "^[^:]+:%d+: CONNECT not sent" )

error_matches( function () cli:ping() end,
               "^[^:]+:%d+: CONNECT not sent" )

error_matches( function () cli:disconnect() end,
               "^[^:]+:%d+: CONNECT not sent" )

truthy( cli:connect({ clean = true }) )

error_matches( function () cli:connect() end,
               "^[^:]+:%d+: CONNECT already sent" )


cli = mqtt({
    socket = sock,
})
cli._send_connect = function (_, _, _, client_id)
    return client_id
end
equals( cli:connect({ clean = true, id = 'foo' }), 'foo' )
equals( cli:connect({ clean = false, id = 'foo' }), 'foo' )
equals( cli:connect({ clean = true, id = '' }), '' )
error_matches( function () cli:connect({ clean = false, id = '' }) end,
               "^[^:]+:%d+: empty Client Identifier" )
matches( cli:connect({ clean = true }), '^LuaMQTT%x+', 'auto-generated' )
error_matches( function () cli:connect({ clean = false }) end,
               "^[^:]+:%d+: missing Client Identifier" )

